package com.hedera.mirror.importer.parser.record.entity;

/*-
 * ‌
 * Hedera Mirror Node
 * ​
 * Copyright (C) 2019 - 2022 Hedera Hashgraph, LLC
 * ​
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * ‍
 */

import static java.lang.String.format;
import static org.assertj.core.api.Assertions.assertThat;
import static org.junit.jupiter.api.Assertions.*;

import com.google.protobuf.BoolValue;
import com.google.protobuf.ByteString;
import com.google.protobuf.Int32Value;
import com.google.protobuf.StringValue;
import com.hederahashgraph.api.proto.java.AccountAmount;
import com.hederahashgraph.api.proto.java.AccountID;
import com.hederahashgraph.api.proto.java.CryptoAddLiveHashTransactionBody;
import com.hederahashgraph.api.proto.java.CryptoCreateTransactionBody;
import com.hederahashgraph.api.proto.java.CryptoDeleteLiveHashTransactionBody;
import com.hederahashgraph.api.proto.java.CryptoUpdateTransactionBody;
import com.hederahashgraph.api.proto.java.Duration;
import com.hederahashgraph.api.proto.java.KeyList;
import com.hederahashgraph.api.proto.java.NftAllowance;
import com.hederahashgraph.api.proto.java.NftRemoveAllowance;
import com.hederahashgraph.api.proto.java.RealmID;
import com.hederahashgraph.api.proto.java.ResponseCodeEnum;
import com.hederahashgraph.api.proto.java.ShardID;
import com.hederahashgraph.api.proto.java.SignedTransaction;
import com.hederahashgraph.api.proto.java.Timestamp;
import com.hederahashgraph.api.proto.java.TokenID;
import com.hederahashgraph.api.proto.java.Transaction;
import com.hederahashgraph.api.proto.java.TransactionBody;
import com.hederahashgraph.api.proto.java.TransactionID;
import com.hederahashgraph.api.proto.java.TransactionRecord;
import com.hederahashgraph.api.proto.java.TransferList;
import java.time.Instant;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Optional;
import java.util.function.Consumer;
import java.util.stream.Collectors;
import java.util.stream.Stream;
import javax.annotation.Resource;
import org.assertj.core.api.Condition;
import org.assertj.core.api.IterableAssert;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.DisplayName;
import org.junit.jupiter.api.Test;
import org.junit.jupiter.params.ParameterizedTest;
import org.junit.jupiter.params.provider.CsvSource;
import org.junit.jupiter.params.provider.EnumSource;

import com.hedera.mirror.common.domain.contract.Contract;
import com.hedera.mirror.common.domain.entity.AbstractEntity;
import com.hedera.mirror.common.domain.entity.Entity;
import com.hedera.mirror.common.domain.entity.EntityId;
import com.hedera.mirror.common.domain.token.Nft;
import com.hedera.mirror.common.domain.token.NftId;
import com.hedera.mirror.common.domain.transaction.CryptoTransfer;
import com.hedera.mirror.common.domain.transaction.ErrataType;
import com.hedera.mirror.common.domain.transaction.LiveHash;
import com.hedera.mirror.common.domain.transaction.NonFeeTransfer;
import com.hedera.mirror.common.domain.transaction.RecordItem;
import com.hedera.mirror.common.domain.transaction.StakingRewardTransfer;
import com.hedera.mirror.common.util.DomainUtils;
import com.hedera.mirror.importer.TestUtils;
import com.hedera.mirror.importer.exception.AliasNotFoundException;
import com.hedera.mirror.importer.parser.PartialDataAction;
import com.hedera.mirror.importer.parser.domain.RecordItemBuilder;
import com.hedera.mirror.importer.parser.record.RecordParserProperties;
import com.hedera.mirror.importer.repository.ContractRepository;
import com.hedera.mirror.importer.repository.CryptoAllowanceRepository;
import com.hedera.mirror.importer.repository.NftAllowanceRepository;
import com.hedera.mirror.importer.repository.NftRepository;
import com.hedera.mirror.importer.repository.TokenAllowanceRepository;
import com.hedera.mirror.importer.util.Utility;
import com.hedera.mirror.importer.util.UtilityTest;

class EntityRecordItemListenerCryptoTest extends AbstractEntityRecordItemListenerTest {
    private static final long INITIAL_BALANCE = 1000L;
    private static final AccountID accountId1 = AccountID.newBuilder().setAccountNum(1001).build();
    private static final long[] additionalTransfers = {5000};
    private static final long[] additionalTransferAmounts = {1001, 1002};
    private static final ByteString ALIAS_KEY = DomainUtils.fromBytes(UtilityTest.ALIAS_ECDSA_SECP256K1);

    @Resource
    private ContractRepository contractRepository;

    @Resource
    private CryptoAllowanceRepository cryptoAllowanceRepository;

    @Resource
    private NftAllowanceRepository nftAllowanceRepository;

    @Resource
    private NftRepository nftRepository;

    @Resource
    private RecordParserProperties parserProperties;

    @Resource
    private TokenAllowanceRepository tokenAllowanceRepository;

    @BeforeEach
    void before() {
        entityProperties.getPersist().setClaims(true);
        entityProperties.getPersist().setCryptoTransferAmounts(true);
        entityProperties.getPersist().setTransactionBytes(false);
    }

    @Test
    void cryptoApproveAllowance() {
        // given
        var consensusTimestamp = recordItemBuilder.timestamp();
        List<Nft> expectedNfts = new LinkedList<>();
        var nftAllowances = customizeNftAllowances(consensusTimestamp, expectedNfts);
        RecordItem recordItem = recordItemBuilder.cryptoApproveAllowance()
                .transactionBody(b -> b.clearNftAllowances().addAllNftAllowances(nftAllowances))
                .record(r -> r.setConsensusTimestamp(consensusTimestamp))
                .build();

        // when
        parseRecordItemAndCommit(recordItem);

        // then
        assertAllowances(recordItem, expectedNfts);
    }

    @Test
    void cryptoCreateWithInitialBalance() {
        Transaction transaction = cryptoCreateTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        CryptoCreateTransactionBody cryptoCreateTransactionBody = transactionBody.getCryptoCreateAccount();
        long initialBalance = cryptoCreateTransactionBody.getInitialBalance();

        var transfer1 = accountAmount(accountId1.getAccountNum(), initialBalance);
        var transfer2 = accountAmount(PAYER.getAccountNum(), -initialBalance);
        TransactionRecord record = transactionRecordSuccess(transactionBody, recordBuilder ->
                groupCryptoTransfersByAccountId(recordBuilder, List.of(transfer1, transfer2)));

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        var accountEntityId = EntityId.of(accountId1);
        var consensusTimestamp = DomainUtils.timeStampInNanos(record.getConsensusTimestamp());
        var dbTransaction = getDbTransaction(record.getConsensusTimestamp());
        Optional<CryptoTransfer> initialBalanceTransfer = cryptoTransferRepository.findById(new CryptoTransfer.Id(
                initialBalance, consensusTimestamp, accountEntityId.getId()));

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(accountEntityId),
                () -> assertCryptoTransfers(4)
                        .areAtMost(1, isAccountAmountReceiverAccountAmount(transfer1.build()))
                        .areAtMost(1, isAccountAmountReceiverAccountAmount(transfer2.build())),
                () -> assertCryptoTransaction(transactionBody, record),
                () -> assertCryptoEntity(cryptoCreateTransactionBody, record.getConsensusTimestamp()),
                () -> assertEquals(cryptoCreateTransactionBody.getInitialBalance(), dbTransaction.getInitialBalance()),
                () -> assertThat(initialBalanceTransfer).isPresent()
        );
    }

    @Test
    void cryptoCreateWithZeroInitialBalance() {
        final long initialBalance = 0;
        CryptoCreateTransactionBody.Builder cryptoCreateBuilder = cryptoCreateAccountBuilderWithDefaults()
                .setInitialBalance(initialBalance);
        Transaction transaction = cryptoCreateTransaction(cryptoCreateBuilder);
        TransactionBody transactionBody = getTransactionBody(transaction);
        CryptoCreateTransactionBody cryptoCreateTransactionBody = transactionBody.getCryptoCreateAccount();
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        var accountEntityId = EntityId.of(accountId1);
        var consensusTimestamp = DomainUtils.timeStampInNanos(record.getConsensusTimestamp());
        var dbTransaction = getDbTransaction(record.getConsensusTimestamp());
        Optional<CryptoTransfer> initialBalanceTransfer = cryptoTransferRepository.findById(new CryptoTransfer.Id(
                initialBalance, consensusTimestamp, accountEntityId.getId()));

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(accountEntityId),
                () -> assertCryptoTransfers(3),
                () -> assertCryptoTransaction(transactionBody, record),
                () -> assertCryptoEntity(cryptoCreateTransactionBody, record.getConsensusTimestamp()),
                () -> assertEquals(cryptoCreateTransactionBody.getInitialBalance(), dbTransaction.getInitialBalance()),
                () -> assertThat(initialBalanceTransfer).isEmpty()
        );
    }

    @Test
    void cryptoCreateFailedTransaction() {
        Transaction transaction = cryptoCreateTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        CryptoCreateTransactionBody cryptoCreateTransactionBody = transactionBody.getCryptoCreateAccount();
        // Clear receipt.accountID since transaction is failure.
        TransactionRecord.Builder recordBuilder = transactionRecord(
                transactionBody, ResponseCodeEnum.INSUFFICIENT_PAYER_BALANCE).toBuilder();
        recordBuilder.getReceiptBuilder().clearAccountID();
        TransactionRecord record = recordBuilder.build();

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        var dbTransaction = getDbTransaction(record.getConsensusTimestamp());

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(),
                () -> assertCryptoTransfers(3),
                () -> assertTransactionAndRecord(transactionBody, record),
                () -> assertNull(dbTransaction.getEntityId()),
                () -> assertEquals(cryptoCreateTransactionBody.getInitialBalance(), dbTransaction.getInitialBalance())
        );
    }

    @Test
    void cryptoCreateInitialBalanceInTransferList() {
        Transaction transaction = cryptoCreateTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        CryptoCreateTransactionBody cryptoCreateTransactionBody = transactionBody.getCryptoCreateAccount();

        // add initial balance to transfer list
        long initialBalance = cryptoCreateTransactionBody.getInitialBalance();
        var transfer1 = accountAmount(accountId1.getAccountNum(), initialBalance);
        var transfer2 = accountAmount(PAYER.getAccountNum(), -initialBalance);
        TransactionRecord record = transactionRecordSuccess(transactionBody, recordBuilder ->
                groupCryptoTransfersByAccountId(recordBuilder, List.of(transfer1, transfer2))
        );

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        var dbTransaction = getDbTransaction(record.getConsensusTimestamp());

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(EntityId.of(accountId1)),
                () -> assertCryptoTransfers(4)
                        .areAtMost(1, isAccountAmountReceiverAccountAmount(transfer1.build()))
                        .areAtMost(1, isAccountAmountReceiverAccountAmount(transfer2.build())),
                () -> assertCryptoTransaction(transactionBody, record),
                () -> assertCryptoEntity(cryptoCreateTransactionBody, record.getConsensusTimestamp()),
                () -> assertEquals(cryptoCreateTransactionBody.getInitialBalance(), dbTransaction.getInitialBalance())
        );
    }

    @Test
    void cryptoCreateAccountAlias() {
        Transaction transaction = cryptoCreateTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        CryptoCreateTransactionBody cryptoCreateTransactionBody = transactionBody.getCryptoCreateAccount();
        TransactionRecord record = buildTransactionRecord(
                recordBuilder -> recordBuilder.setAlias(ALIAS_KEY).getReceiptBuilder().setAccountID(accountId1),
                transactionBody,
                ResponseCodeEnum.SUCCESS.getNumber());

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        var accountEntityId = EntityId.of(accountId1);
        var consensusTimestamp = DomainUtils.timeStampInNanos(record.getConsensusTimestamp());
        var dbTransaction = getDbTransaction(record.getConsensusTimestamp());
        Optional<CryptoTransfer> initialBalanceTransfer = cryptoTransferRepository.findById(new CryptoTransfer.Id(
                0, consensusTimestamp, accountEntityId.getId()));

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(accountEntityId),
                () -> assertCryptoTransfers(3),
                () -> assertCryptoTransaction(transactionBody, record),
                () -> assertCryptoEntity(cryptoCreateTransactionBody, record.getConsensusTimestamp()),
                () -> assertEquals(cryptoCreateTransactionBody.getInitialBalance(), dbTransaction.getInitialBalance()),
                () -> assertThat(initialBalanceTransfer).isEmpty(),
                () -> assertThat(entityRepository.findByAlias(ALIAS_KEY.toByteArray())).get()
                        .isEqualTo(accountEntityId.getId())
        );
    }

    @Test
    void cryptoDeleteAllowance() {
        // given
        var delegatingSpender = EntityId.of(recordItemBuilder.accountId());
        var ownerAccountId = EntityId.of(recordItemBuilder.accountId());
        var spender1 = EntityId.of(recordItemBuilder.accountId());
        var spender2 = EntityId.of(recordItemBuilder.accountId());
        var tokenId1 = EntityId.of(recordItemBuilder.tokenId());
        var tokenId2 = EntityId.of(recordItemBuilder.tokenId());
        List<NftRemoveAllowance> nftRemoveAllowances = List.of(
                NftRemoveAllowance.newBuilder()
                        .setOwner(AccountID.newBuilder().setAccountNum(ownerAccountId.getEntityNum()))
                        .setTokenId(TokenID.newBuilder().setTokenNum(tokenId1.getEntityNum()))
                        .addSerialNumbers(1L)
                        .addSerialNumbers(2L)
                        .build(),
                NftRemoveAllowance.newBuilder()
                        .setOwner(AccountID.newBuilder().setAccountNum(ownerAccountId.getEntityNum()))
                        .setTokenId(TokenID.newBuilder().setTokenNum(tokenId2.getEntityNum()))
                        .addSerialNumbers(1L)
                        .addSerialNumbers(2L)
                        .addSerialNumbers(2L)
                        .build()
        );
        RecordItem recordItem = recordItemBuilder.cryptoDeleteAllowance()
                .transactionBody(b -> b.clearNftAllowances().addAllNftAllowances(nftRemoveAllowances))
                .build();

        var nft1 = Nft.builder()
                .accountId(ownerAccountId)
                .createdTimestamp(10L)
                .deleted(false)
                .id(new NftId(1L, tokenId1))
                .modifiedTimestamp(recordItem.getConsensusTimestamp())
                .build();
        var nft2 = Nft.builder()
                .accountId(ownerAccountId)
                .createdTimestamp(11L)
                .deleted(false)
                .id(new NftId(2L, tokenId1))
                .modifiedTimestamp(recordItem.getConsensusTimestamp())
                .build();
        var nft3 = Nft.builder()
                .accountId(ownerAccountId)
                .createdTimestamp(12L)
                .deleted(false)
                .id(new NftId(1L, tokenId2))
                .modifiedTimestamp(recordItem.getConsensusTimestamp())
                .build();
        var nft4 = Nft.builder()
                .accountId(ownerAccountId)
                .createdTimestamp(13L)
                .deleted(false)
                .id(new NftId(2L, tokenId2))
                .modifiedTimestamp(recordItem.getConsensusTimestamp())
                .build();
        List<Nft> nftsWithAllowance = Stream.of(
                nft1.toBuilder().delegatingSpender(delegatingSpender).modifiedTimestamp(15L).spender(spender1),
                nft2.toBuilder().modifiedTimestamp(16L).spender(spender2),
                nft3.toBuilder().modifiedTimestamp(17L).spender(spender1),
                nft4.toBuilder().modifiedTimestamp(18L).spender(spender2)
        ).map(Nft.NftBuilder::build).collect(Collectors.toList());
        nftRepository.saveAll(nftsWithAllowance);

        // when
        parseRecordItemAndCommit(recordItem);

        // then
        assertAll(
                () -> assertEquals(0, entityRepository.count()),
                () -> assertTransactionAndRecord(recordItem.getTransactionBody(), recordItem.getRecord()),
                () -> assertThat(nftRepository.findAll()).containsExactlyInAnyOrder(nft1, nft2, nft3, nft4)
        );
    }

    @Test
    void cryptoUpdateSuccessfulTransaction() {
        createAccount();

        // now update
        Transaction transaction = cryptoUpdateTransaction(accountId1);
        TransactionBody transactionBody = getTransactionBody(transaction);
        CryptoUpdateTransactionBody cryptoUpdateTransactionBody = transactionBody.getCryptoUpdateAccount();
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        parseRecordItemAndCommit(new RecordItem(transaction, record));
        Entity dbAccountEntity = getTransactionEntity(record.getConsensusTimestamp());

        assertAll(
                () -> assertEquals(2, transactionRepository.count()),
                () -> assertEntities(EntityId.of(accountId1)),
                () -> assertEquals(6, cryptoTransferRepository.count()),
                () -> assertCryptoTransaction(transactionBody, record),

                // transaction body inputs
                () -> assertEquals(cryptoUpdateTransactionBody.getAutoRenewPeriod().getSeconds(),
                        dbAccountEntity.getAutoRenewPeriod()),
                () -> assertEquals(DomainUtils.getPublicKey(
                        cryptoUpdateTransactionBody.getKey().toByteArray()), dbAccountEntity.getPublicKey()),
                () -> assertEquals(EntityId.of(cryptoUpdateTransactionBody.getProxyAccountID()),
                        dbAccountEntity.getProxyAccountId()),
                () -> assertArrayEquals(cryptoUpdateTransactionBody.getKey()
                        .toByteArray(), dbAccountEntity.getKey()),
                () -> assertEquals(cryptoUpdateTransactionBody.getMaxAutomaticTokenAssociations().getValue(),
                        dbAccountEntity.getMaxAutomaticTokenAssociations()),
                () -> assertEquals(cryptoUpdateTransactionBody.getMemo().getValue(), dbAccountEntity.getMemo()),
                () -> assertEquals(DomainUtils.timeStampInNanos(cryptoUpdateTransactionBody.getExpirationTime()),
                        dbAccountEntity.getExpirationTimestamp()),
                () -> assertEquals(DomainUtils.timestampInNanosMax(record.getConsensusTimestamp()),
                        dbAccountEntity.getTimestampLower()),
                () -> assertFalse(dbAccountEntity.getReceiverSigRequired()),
                () -> assertFalse(dbAccountEntity.getDeclineReward()),
                () -> assertEquals(cryptoUpdateTransactionBody.getStakedNodeId(), dbAccountEntity.getStakedNodeId()),
                () -> assertEquals(AbstractEntity.ACCOUNT_ID_CLEARED, dbAccountEntity.getStakedAccountId()),
                () -> assertEquals(Utility.getEpochDay(DomainUtils.timestampInNanosMax(record.getConsensusTimestamp())),
                        dbAccountEntity.getStakePeriodStart())
        );
    }

    @Test
    void cryptoTransferWithPaidStakingRewards() {
        // given
        var receiver1 = domainBuilder.entity().persist();
        var receiver2 = domainBuilder.entity().persist();
        var sender = domainBuilder.entity().persist();
        var receiver1Id = AccountID.newBuilder().setAccountNum(receiver1.getNum()).build();
        var receiver2Id = AccountID.newBuilder().setAccountNum(receiver2.getNum()).build();
        var senderId = AccountID.newBuilder().setAccountNum(sender.getNum()).build();
        var timestamp = 1653506322000111222L; // 19137 days since epoch
        var stakePeriodStart = 19137L;

        var recordItem = recordItemBuilder.cryptoTransfer()
                .transactionBody(b -> b.setTransfers(TransferList.newBuilder()
                        .addAccountAmounts(AccountAmount.newBuilder().setAccountID(senderId).setAmount(-20L))
                        .addAccountAmounts(AccountAmount.newBuilder().setAccountID(receiver1Id).setAmount(5L))
                        .addAccountAmounts(AccountAmount.newBuilder().setAccountID(receiver2Id).setAmount(-15L))))
                .record(r -> {
                    // preserve the tx fee paid by the payer and received by the node and the fee collector
                    // sender only gets deducted 15, since it gets a 5 reward payout
                    // receiver1 gets 9, since it gets a 4 reward payout
                    // receiver2 gets no reward
                    var paidStakingRewards = List.of(
                            AccountAmount.newBuilder().setAccountID(senderId).setAmount(5L).build(),
                            AccountAmount.newBuilder().setAccountID(receiver1Id).setAmount(4L).build()
                    );

                    var transferList = r.getTransferList().toBuilder()
                            .addAccountAmounts(AccountAmount.newBuilder().setAccountID(senderId).setAmount(-15L))
                            .addAccountAmounts(AccountAmount.newBuilder().setAccountID(receiver1Id).setAmount(9L))
                            .addAccountAmounts(AccountAmount.newBuilder().setAccountID(receiver2Id).setAmount(15L))
                            .addAccountAmounts(AccountAmount.newBuilder()
                                    .setAccountID(RecordItemBuilder.STAKING_REWARD_ACCOUNT).setAmount(-9L));
                    r.setConsensusTimestamp(TestUtils.toTimestamp(timestamp))
                            .setTransferList(transferList)
                            .addAllPaidStakingRewards(paidStakingRewards);
                })
                .build();

        // when
        parseRecordItemAndCommit(recordItem);

        // then
        sender.setStakePeriodStart(stakePeriodStart);
        receiver1.setStakePeriodStart(stakePeriodStart);

        var payerAccountId = recordItem.getPayerAccountId();
        var expectedStakingRewardTransfer1 = new StakingRewardTransfer();
        expectedStakingRewardTransfer1.setAccountId(sender.getId());
        expectedStakingRewardTransfer1.setAmount(5L);
        expectedStakingRewardTransfer1.setConsensusTimestamp(timestamp);
        expectedStakingRewardTransfer1.setPayerAccountId(payerAccountId);
        var expectedStakingRewardTransfer2 = new StakingRewardTransfer();
        expectedStakingRewardTransfer2.setAccountId(receiver1.getId());
        expectedStakingRewardTransfer2.setAmount(4L);
        expectedStakingRewardTransfer2.setConsensusTimestamp(timestamp);
        expectedStakingRewardTransfer2.setPayerAccountId(payerAccountId);

        assertAll(
                () -> assertEquals(0, contractRepository.count()),
                // 3 for fee, 3 for hbar transfers, and 1 for reward payout from 0.0.800
                () -> assertEquals(7, cryptoTransferRepository.count()),
                () -> assertThat(entityRepository.findAll()).containsExactlyInAnyOrder(sender, receiver1, receiver2),
                () -> assertThat(stakingRewardTransferRepository.findAll()).containsExactlyInAnyOrder(
                        expectedStakingRewardTransfer1, expectedStakingRewardTransfer2
                ),
                () -> assertEquals(1, transactionRepository.count())
        );
    }

    @Test
    void cryptoTransferFailedWithPaidStakingRewards() {
        // given
        var payer = domainBuilder.entity().persist();
        var receiver = domainBuilder.entity().persist();
        var sender = domainBuilder.entity().persist();
        var payerId = AccountID.newBuilder().setAccountNum(payer.getNum()).build();
        var receiverId = AccountID.newBuilder().setAccountNum(receiver.getNum()).build();
        var senderId = AccountID.newBuilder().setAccountNum(sender.getNum()).build();
        var timestamp = 1653506322000111222L; // 19137 days since epoch
        var stakePeriodStart = 19137L;

        // Transaction failed with INSUFFICIENT_ACCOUNT_BALANCE because sender's balance is less than the indented
        // transfer amount. However, the transaction payer has a balance change and there is pending reward for the
        // payer
        // account, so there will be a reward payout for the transaction payer.
        var transactionId = TransactionID.newBuilder().setAccountID(payerId)
                .setTransactionValidStart(TestUtils.toTimestamp(timestamp - 200L)).build();
        var recordItem = recordItemBuilder.cryptoTransfer()
                .transactionBody(b -> b.setTransfers(TransferList.newBuilder()
                        .addAccountAmounts(AccountAmount.newBuilder().setAccountID(senderId).setAmount(-20L))
                        .addAccountAmounts(AccountAmount.newBuilder().setAccountID(receiverId).setAmount(20L))))
                .transactionBodyWrapper(b -> b.setTransactionID(transactionId))
                .record(r -> r.setConsensusTimestamp(TestUtils.toTimestamp(timestamp))
                        .setTransactionID(transactionId)
                        .setTransferList(TransferList.newBuilder()
                                .addAccountAmounts(recordItemBuilder.accountAmount(payerId, -2800L))
                                .addAccountAmounts(recordItemBuilder.accountAmount(NODE, 1000L))
                                .addAccountAmounts(recordItemBuilder.accountAmount(TREASURY, 2000L))
                                .addAccountAmounts(recordItemBuilder.accountAmount(
                                        RecordItemBuilder.STAKING_REWARD_ACCOUNT, -200L)))
                        .addPaidStakingRewards(recordItemBuilder.accountAmount(payerId, 200L))
                        .getReceiptBuilder().setStatus(ResponseCodeEnum.INSUFFICIENT_ACCOUNT_BALANCE))
                .build();

        // when
        parseRecordItemAndCommit(recordItem);

        // then
        payer.setStakePeriodStart(stakePeriodStart);

        var expectedStakingRewardTransfer = new StakingRewardTransfer();
        expectedStakingRewardTransfer.setAccountId(payer.getId());
        expectedStakingRewardTransfer.setAmount(200L);
        expectedStakingRewardTransfer.setConsensusTimestamp(timestamp);
        expectedStakingRewardTransfer.setPayerAccountId(payer.toEntityId());

        assertAll(
                () -> assertEquals(0, contractRepository.count()),
                () -> assertEquals(4, cryptoTransferRepository.count()),
                () -> assertThat(entityRepository.findAll()).containsExactlyInAnyOrder(payer, sender, receiver),
                () -> assertThat(stakingRewardTransferRepository.findAll()).containsOnly(expectedStakingRewardTransfer),
                () -> assertEquals(1, transactionRepository.count())
        );
    }

    /**
     * Github issue #483
     */
    @Test
    void samePayerAndUpdateAccount() {
        Transaction transaction = cryptoUpdateTransaction(accountId1);
        TransactionBody transactionBody = getTransactionBody(transaction);
        transactionBody = TransactionBody.newBuilder()
                .mergeFrom(transactionBody)
                .setTransactionID(Utility.getTransactionId(accountId1))
                .build();
        transaction = Transaction.newBuilder()
                .setSignedTransactionBytes(SignedTransaction.newBuilder()
                        .setBodyBytes(transactionBody.toByteString())
                        .build().toByteString())
                .build();
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        assertThat(transactionRepository.findById(DomainUtils.timestampInNanosMax(record.getConsensusTimestamp())))
                .get()
                .extracting(com.hedera.mirror.common.domain.transaction.Transaction::getPayerAccountId,
                        com.hedera.mirror.common.domain.transaction.Transaction::getEntityId)
                .containsOnly(EntityId.of(accountId1));
    }

    // Transactions in production have proxyAccountID explicitly set to '0.0.0'. Test is to prevent code regression
    // in handling this weird case.
    @Test
    void proxyAccountIdSetTo0() {
        // given
        Transaction transaction = cryptoUpdateTransaction(accountId1);
        TransactionBody transactionBody = getTransactionBody(transaction);
        var bodyBuilder = transactionBody.toBuilder();
        bodyBuilder.getCryptoUpdateAccountBuilder().setProxyAccountID(AccountID.getDefaultInstance());
        transactionBody = bodyBuilder.build();
        transaction = Transaction.newBuilder().setSignedTransactionBytes(SignedTransaction.newBuilder()
                        .setBodyBytes(transactionBody.toByteString())
                        .build().toByteString())
                .build();
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        // then: process the transaction without throwing NPE
        parseRecordItemAndCommit(new RecordItem(transaction, record));

        assertThat(transactionRepository.count()).isEqualTo(1L);
        assertThat(entityRepository.findById(EntityId.of(accountId1).getId()))
                .get()
                .extracting(Entity::getProxyAccountId)
                .isNull();
    }

    @DisplayName("update account such that expiration timestamp overflows nanos_timestamp")
    @ParameterizedTest(name = "with seconds {0} and expectedNanosTimestamp {1}")
    @CsvSource({
            "9223372036854775807, 9223372036854775807",
            "31556889864403199, 9223372036854775807",
            "-9223372036854775808, -9223372036854775808",
            "-1000000000000000000, -9223372036854775808"
    })
    void cryptoUpdateExpirationOverflow(long seconds, long expectedNanosTimestamp) {
        createAccount();

        // now update
        var updateTransaction = buildTransaction(builder -> {
            builder.getCryptoUpdateAccountBuilder()
                    .setAccountIDToUpdate(accountId1)
                    // *** THIS IS THE OVERFLOW WE WANT TO TEST ***
                    // This should result in the entity having a Long.MAX_VALUE or Long.MIN_VALUE expirations
                    // (the results of overflows).
                    .setExpirationTime(Timestamp.newBuilder().setSeconds(seconds))
                    .setDeclineReward(BoolValue.of(true))
                    .setStakedAccountId(AccountID.newBuilder().setAccountNum(1L).build());
        });
        var transactionBody = getTransactionBody(updateTransaction);

        var record = transactionRecordSuccess(transactionBody);

        parseRecordItemAndCommit(new RecordItem(updateTransaction, record));

        var dbAccountEntity = getTransactionEntity(record.getConsensusTimestamp());
        var stakedAccountId = EntityId.of(transactionBody.getCryptoUpdateAccount().getStakedAccountId()).getId();

        assertAll(
                () -> assertEquals(2, transactionRepository.count()),
                () -> assertEquals(expectedNanosTimestamp, dbAccountEntity.getExpirationTimestamp()),
                () -> assertTrue(dbAccountEntity.getDeclineReward()),
                () -> assertEquals(stakedAccountId, dbAccountEntity.getStakedAccountId()),
                () -> assertEquals(-1L, dbAccountEntity.getStakedNodeId())
        );
    }

    @Test
    void cryptoUpdateFailedTransaction() {
        Transaction createTransaction = cryptoCreateTransaction();
        TransactionRecord createRecord = transactionRecordSuccess(
                getTransactionBody(createTransaction));
        parseRecordItemAndCommit(new RecordItem(createTransaction, createRecord));

        // now update
        Transaction transaction = cryptoUpdateTransaction(accountId1);
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = transactionRecord(transactionBody, ResponseCodeEnum.INSUFFICIENT_ACCOUNT_BALANCE);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        Entity dbAccountEntityBefore = getTransactionEntity(createRecord.getConsensusTimestamp());
        Entity dbAccountEntity = getTransactionEntity(record.getConsensusTimestamp());

        assertAll(
                () -> assertEquals(2, transactionRepository.count()),
                () -> assertEntities(EntityId.of(accountId1)),
                () -> assertCryptoTransfers(6), // 3 + 3 fee transfers with one transfer per account
                () -> assertTransactionAndRecord(transactionBody, record),
                () -> assertAccount(record.getReceipt().getAccountID(), dbAccountEntity),
                () -> assertEquals(dbAccountEntityBefore, dbAccountEntity)// no changes to entity
        );
    }

    @Test
    void cryptoDeleteSuccessfulTransaction() {
        // first create the account
        createAccount();
        Entity dbAccountEntityBefore = getEntity(EntityId.of(accountId1));

        // now delete
        Transaction transaction = cryptoDeleteTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        Entity dbAccountEntity = getTransactionEntity(record.getConsensusTimestamp());

        assertAll(
                () -> assertEquals(2, transactionRepository.count()),
                () -> assertEntities(EntityId.of(accountId1)),
                () -> assertCryptoTransfers(6), // 3 + 3 fee transfers with one transfer per account
                () -> assertCryptoTransaction(transactionBody, record),
                () -> assertThat(dbAccountEntity)
                        .isNotNull()
                        .returns(true, Entity::getDeleted)
                        .returns(DomainUtils.timestampInNanosMax(record.getConsensusTimestamp()),
                                Entity::getTimestampLower)
                        .usingRecursiveComparison()
                        .ignoringFields("deleted", "timestampRange")
                        .isEqualTo(dbAccountEntityBefore)
        );
    }

    @Test
    void cryptoDeleteFailedTransaction() {
        createAccount();

        // now delete
        Transaction transaction = cryptoDeleteTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = transactionRecord(transactionBody,
                ResponseCodeEnum.INSUFFICIENT_ACCOUNT_BALANCE.getNumber(),
                recordBuilder -> groupCryptoTransfersByAccountId(recordBuilder, List.of()));

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        Entity dbAccountEntity = getTransactionEntity(record.getConsensusTimestamp());

        assertAll(
                () -> assertEquals(2, transactionRepository.count()),
                () -> assertEntities(EntityId.of(accountId1)),
                () -> assertCryptoTransfers(6), // 3 + 3 fee transfers with only one transfer per account
                () -> assertCryptoTransaction(transactionBody, record),
                () -> assertThat(dbAccountEntity)
                        .isNotNull()
                        .returns(false, Entity::getDeleted)
        );
    }

    @Test
    void cryptoAddLiveHashPersist() {
        Transaction transaction = cryptoAddLiveHashTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        CryptoAddLiveHashTransactionBody cryptoAddLiveHashTransactionBody = transactionBody.getCryptoAddLiveHash();
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        var dbTransaction = getDbTransaction(record.getConsensusTimestamp());
        LiveHash dbLiveHash = liveHashRepository.findById(dbTransaction.getConsensusTimestamp()).get();

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(),
                () -> assertCryptoTransfers(3),
                () -> assertEquals(1, liveHashRepository.count()),
                () -> assertTransactionAndRecord(transactionBody, record),
                () -> assertArrayEquals(cryptoAddLiveHashTransactionBody.getLiveHash().getHash().toByteArray(),
                        dbLiveHash.getLivehash())
        );
    }

    @Test
    void cryptoAddLiveHashDoNotPersist() {
        entityProperties.getPersist().setClaims(false);
        Transaction transaction = cryptoAddLiveHashTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(),
                () -> assertCryptoTransfers(3),
                () -> assertEquals(0, liveHashRepository.count()),
                () -> assertTransactionAndRecord(transactionBody, record)
        );
    }

    @Test
    void cryptoDeleteLiveHash() {
        Transaction transactionAddLiveHash = cryptoAddLiveHashTransaction();
        parseRecordItemAndCommit(new RecordItem(transactionAddLiveHash,
                transactionRecordSuccess(getTransactionBody(transactionAddLiveHash))));

        // now delete the live hash
        Transaction transaction = cryptoDeleteLiveHashTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        CryptoDeleteLiveHashTransactionBody deleteLiveHashTransactionBody = transactionBody.getCryptoDeleteLiveHash();
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        assertAll(
                () -> assertEquals(2, transactionRepository.count()),
                () -> assertEntities(),
                () -> assertCryptoTransfers(6),
                () -> assertEquals(1, liveHashRepository.count()),
                () -> assertTransactionAndRecord(transactionBody, record)
        );
    }

    @Test
    void cryptoTransferWithPersistence() {
        entityProperties.getPersist().setCryptoTransferAmounts(true);
        // make the transfers
        Transaction transaction = cryptoTransferTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(),
                () -> assertEquals(4, cryptoTransferRepository.count()),
                () -> assertTransactionAndRecord(transactionBody, record)
        );
    }

    @Test
    void cryptoTransferWithoutPersistence() {
        entityProperties.getPersist().setCryptoTransferAmounts(false);
        // make the transfers
        Transaction transaction = cryptoTransferTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(),
                () -> assertEquals(0, entityRepository.count()),
                () -> assertCryptoTransfers(0),
                () -> assertTransactionAndRecord(transactionBody, record)
        );
    }

    @Test
    void cryptoTransferFailedTransaction() {
        entityProperties.getPersist().setCryptoTransferAmounts(true);
        // make the transfers
        Transaction transaction = cryptoTransferTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = transactionRecord(transactionBody, ResponseCodeEnum.INVALID_ACCOUNT_ID);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(),
                () -> assertCryptoTransfers(3),
                () -> assertEquals(0, nonFeeTransferRepository.count()),
                () -> assertTransactionAndRecord(transactionBody, record)
        );
    }

    @Test
    void cryptoTransferFailedTransactionErrata() {
        entityProperties.getPersist().setCryptoTransferAmounts(true);
        Transaction transaction = cryptoTransferTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = buildTransactionRecord(r -> {
            for (int i = 0; i < additionalTransfers.length; i++) {
                // Add non-fee transfers to record
                var accountAmount = accountAmount(additionalTransfers[i], additionalTransferAmounts[i]);
                r.getTransferListBuilder().addAccountAmounts(accountAmount);
            }
        }, transactionBody, ResponseCodeEnum.INVALID_ACCOUNT_ID.getNumber());

        var recordItem = new RecordItem(transaction, record);
        parseRecordItemAndCommit(recordItem);

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEntities(),
                () -> assertEquals(4, cryptoTransferRepository.count(), "Node, network fee & errata"),
                () -> assertEquals(0, nonFeeTransferRepository.count()),
                () -> assertTransactionAndRecord(transactionBody, record),
                () -> {
                    for (int i = 0; i < additionalTransfers.length; i++) {
                        var id = new CryptoTransfer.Id(additionalTransferAmounts[i],
                                recordItem.getConsensusTimestamp(), additionalTransfers[i]);
                        assertThat(cryptoTransferRepository.findById(id))
                                .get()
                                .extracting(CryptoTransfer::getErrata)
                                .isEqualTo(ErrataType.DELETE);
                    }
                }
        );
    }

    @Test
    void cryptoTransferHasCorrectIsApprovalValue() {
        final long[] accountNums = {PAYER.getAccountNum(), PAYER2.getAccountNum(), PAYER3.getAccountNum()};
        final long[] amounts = {210, -300, 15};
        final boolean[] isApprovals = {false, true, false};
        Transaction transaction = buildTransaction(r -> {
            for (int i = 0; i < accountNums.length; i++) {
                var accountAmount = accountAmount(accountNums[i], amounts[i]).setIsApproval(isApprovals[i]).build();
                r.getCryptoTransferBuilder()
                        .getTransfersBuilder()
                        .addAccountAmounts(accountAmount);
            }
        });
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = buildTransactionRecordWithNoTransactions(builder -> {
            for (int i = 0; i < accountNums.length; i++) {
                var accountAmount = accountAmount(accountNums[i], amounts[i]).setIsApproval(false).build();
                builder.getTransferListBuilder()
                        .addAccountAmounts(accountAmount);
            }
        }, transactionBody, ResponseCodeEnum.SUCCESS.getNumber());

        var recordItem = new RecordItem(transaction, record);
        parseRecordItemAndCommit(recordItem);

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEquals(amounts.length, cryptoTransferRepository.count()),
                () -> {
                    for (var cryptoTransfer : cryptoTransferRepository.findAll()) {
                        for (int i = 0; i < isApprovals.length; i++) {
                            if (cryptoTransfer.getEntityId() != accountNums[i]) {
                                continue;
                            }
                            assertThat(cryptoTransfer)
                                    .extracting(CryptoTransfer::getIsApproval)
                                    .isEqualTo(isApprovals[i]);
                        }
                    }
                }
        );
    }

    @Test
    void cryptoTransferWithAlias() {
        entityProperties.getPersist().setCryptoTransferAmounts(true);
        entityProperties.getPersist().setNonFeeTransfers(true);
        Entity entity = domainBuilder.entity().persist();
        var newAccount = AccountID.newBuilder().setAccountNum(domainBuilder.id()).build();
        assertThat(entityRepository.findByAlias(entity.getAlias())).get().isEqualTo(entity.getId());
        assertThat(entityRepository.findByAlias(ALIAS_KEY.toByteArray())).isNotPresent();

        // Crypto create alias account
        Transaction accountCreateTransaction = cryptoCreateTransaction();
        TransactionBody accountCreateTransactionBody = getTransactionBody(accountCreateTransaction);
        TransactionRecord recordCreate = buildTransactionRecord(
                recordBuilder -> recordBuilder.setAlias(ALIAS_KEY).getReceiptBuilder().setAccountID(newAccount),
                accountCreateTransactionBody,
                ResponseCodeEnum.SUCCESS.getNumber());

        var transfer1 = accountAliasAmount(ALIAS_KEY, 1003).build();
        var transfer2 = accountAliasAmount(ByteString.copyFrom(entity.getAlias()), 1004).build();
        // Crypto transfer to both existing alias and newly created alias accounts
        Transaction transaction = buildTransaction(builder -> builder.getCryptoTransferBuilder().getTransfersBuilder()
                .addAccountAmounts(transfer1)
                .addAccountAmounts(transfer2));
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord recordTransfer = transactionRecordSuccess(transactionBody,
                builder -> groupCryptoTransfersByAccountId(builder, List.of()));

        parseRecordItemsAndCommit(List.of(new RecordItem(accountCreateTransaction, recordCreate),
                new RecordItem(transaction, recordTransfer)));

        assertAll(
                () -> assertEquals(2, transactionRepository.count()),
                () -> assertEntities(EntityId.of(newAccount), entity.toEntityId()),
                () -> assertCryptoTransfers(6)
                        .areAtMost(1, isAccountAmountReceiverAccountAmount(transfer1))
                        .areAtMost(1, isAccountAmountReceiverAccountAmount(transfer2)),
                () -> assertEquals(additionalTransfers.length * 2 + 2, nonFeeTransferRepository.count()),
                () -> assertTransactionAndRecord(transactionBody, recordTransfer),
                () -> assertThat(findNonFeeTransfers())
                        .extracting(NonFeeTransfer::getEntityId)
                        .extracting(EntityId::getEntityNum)
                        .contains(newAccount.getAccountNum(), entity.getNum())
        );
    }

    @Test
    void cryptoTransferWithEvmAddressAlias() {
        Contract contract = domainBuilder.contract().persist();
        assertThat(contractRepository.findByEvmAddress(contract.getEvmAddress())).isPresent();

        entityProperties.getPersist().setNonFeeTransfers(true);

        long transferAmount = 123;
        var transfer1 = accountAliasAmount(DomainUtils.fromBytes(contract.getEvmAddress()), transferAmount).build();
        Transaction transaction = buildTransaction(builder -> builder.getCryptoTransferBuilder().getTransfersBuilder()
                .addAccountAmounts(transfer1));
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord transactionRecord = transactionRecordSuccess(transactionBody);
        RecordItem recordItem = new RecordItem(transaction, transactionRecord);
        parseRecordItemAndCommit(recordItem);

        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEquals(1, nonFeeTransferRepository.count()),
                () -> assertTransactionAndRecord(transactionBody, transactionRecord),
                () -> assertThat(findNonFeeTransfers())
                        .allSatisfy(nonFeeTransfer -> {
                            assertThat(nonFeeTransfer.getEntityId()).isEqualTo(contract.toEntityId());
                            assertThat(nonFeeTransfer.getAmount()).isEqualTo(transferAmount);
                            assertThat(nonFeeTransfer.getPayerAccountId()).isEqualTo(recordItem.getPayerAccountId());
                        })
        );
    }

    private Condition<CryptoTransfer> isAccountAmountReceiverAccountAmount(AccountAmount receiver) {
        return new Condition<>(
                cryptoTransfer ->
                        isAccountAmountReceiverAccountAmount(cryptoTransfer, receiver),
                format("Is %s the receiver account amount.", receiver));
    }

    @ParameterizedTest
    @EnumSource(value = PartialDataAction.class, names = {"DEFAULT", "SKIP"})
    void cryptoTransferWithUnknownAlias(PartialDataAction partialDataAction) {
        // given
        // both accounts have alias, and only account2's alias is in db
        entityProperties.getPersist().setCryptoTransferAmounts(true);
        entityProperties.getPersist().setNonFeeTransfers(true);
        parserProperties.setPartialDataAction(partialDataAction);

        Entity account1 = domainBuilder.entity().get();
        Entity account2 = domainBuilder.entity().persist();

        // crypto transfer from unknown account1 alias to account2 alias
        Transaction transaction = buildTransaction(builder -> builder.getCryptoTransferBuilder().getTransfersBuilder()
                .addAccountAmounts(accountAliasAmount(DomainUtils.fromBytes(account1.getAlias()), 100))
                .addAccountAmounts(accountAliasAmount(DomainUtils.fromBytes(account2.getAlias()), -100)));
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord transactionRecord = buildTransactionRecord(r -> r.getTransferListBuilder()
                        .addAccountAmounts(accountAmount(account1.getNum(), 100))
                        .addAccountAmounts(accountAmount(account2.getNum(), -100)),
                transactionBody, ResponseCodeEnum.SUCCESS.getNumber());
        List<EntityId> expectedEntityIds = partialDataAction == PartialDataAction.DEFAULT ?
                Arrays.asList(account2.toEntityId(), null) : List.of(account2.toEntityId());

        // when
        parseRecordItemAndCommit(new RecordItem(transaction, transactionRecord));

        // then
        assertAll(
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertEquals(5, cryptoTransferRepository.count()),
                () -> assertTransactionAndRecord(transactionBody, transactionRecord),
                () -> assertThat(findNonFeeTransfers())
                        .extracting(NonFeeTransfer::getEntityId)
                        .containsExactlyInAnyOrderElementsOf(expectedEntityIds)
        );
    }

    @Test
    void cryptoTransferWithUnknownAliasActionError() {
        // given
        // both accounts have alias, and only account2's alias is in db
        entityProperties.getPersist().setCryptoTransferAmounts(true);
        entityProperties.getPersist().setNonFeeTransfers(true);
        parserProperties.setPartialDataAction(PartialDataAction.ERROR);

        Entity account1 = domainBuilder.entity().get();
        Entity account2 = domainBuilder.entity().persist();

        // crypto transfer from unknown account1 alias to account2 alias
        Transaction transaction = buildTransaction(builder -> builder.getCryptoTransferBuilder().getTransfersBuilder()
                .addAccountAmounts(accountAliasAmount(DomainUtils.fromBytes(account1.getAlias()), 100))
                .addAccountAmounts(accountAliasAmount(DomainUtils.fromBytes(account2.getAlias()), -100)));
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord transactionRecord = buildTransactionRecord(r -> r.getTransferListBuilder()
                        .addAccountAmounts(accountAmount(account1.getNum(), 100))
                        .addAccountAmounts(accountAmount(account2.getNum(), -100)),
                transactionBody, ResponseCodeEnum.SUCCESS.getNumber());

        // when, then
        assertThrows(AliasNotFoundException.class,
                () -> parseRecordItemAndCommit(new RecordItem(transaction, transactionRecord)));
        assertAll(
                () -> assertEquals(0, transactionRepository.count()),
                () -> assertEquals(0, cryptoTransferRepository.count()),
                () -> assertThat(findNonFeeTransfers()).isEmpty()
        );
    }

    @Test
    void unknownTransactionResult() {
        int unknownResult = -1000;
        Transaction transaction = cryptoCreateTransaction();
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = transactionRecord(transactionBody, unknownResult);

        parseRecordItemAndCommit(new RecordItem(transaction, record));

        assertThat(transactionRepository.findAll())
                .hasSize(1)
                .extracting(com.hedera.mirror.common.domain.transaction.Transaction::getResult)
                .containsOnly(unknownResult);
    }

    @Test
    void cryptoTransferPersistRawBytesDefault() {
        // Use the default properties for record parsing - the raw bytes should NOT be stored in the db
        Transaction transaction = cryptoTransferTransaction();
        testRawBytes(transaction, null);
    }

    @Test
    void cryptoTransferPersistRawBytesTrue() {
        // Explicitly persist the transaction bytes
        entityProperties.getPersist().setTransactionBytes(true);
        Transaction transaction = cryptoTransferTransaction();
        testRawBytes(transaction, transaction.toByteArray());
    }

    @Test
    void cryptoTransferPersistRawBytesFalse() {
        // Explicitly DO NOT persist the transaction bytes
        entityProperties.getPersist().setTransactionBytes(false);
        Transaction transaction = cryptoTransferTransaction();
        testRawBytes(transaction, null);
    }

    private void assertAllowances(RecordItem recordItem, Collection<Nft> expectedNfts) {
        assertAll(
                () -> assertEquals(1, cryptoAllowanceRepository.count()),
                () -> assertEquals(3, cryptoTransferRepository.count()),
                () -> assertEquals(0, entityRepository.count()),
                () -> assertEquals(3, nftAllowanceRepository.count()),
                () -> assertEquals(1, tokenAllowanceRepository.count()),
                () -> assertEquals(1, transactionRepository.count()),
                () -> assertTransactionAndRecord(recordItem.getTransactionBody(), recordItem.getRecord()),
                () -> assertThat(cryptoAllowanceRepository.findAll())
                        .allSatisfy(a -> assertThat(a.getAmount()).isPositive())
                        .allSatisfy(a -> assertThat(a.getOwner()).isPositive())
                        .allSatisfy(a -> assertThat(a.getSpender()).isPositive())
                        .allMatch(a -> recordItem.getConsensusTimestamp() == a.getTimestampLower())
                        .allMatch(a -> recordItem.getPayerAccountId().equals(a.getPayerAccountId())),
                () -> assertThat(nftAllowanceRepository.findAll())
                        .allSatisfy(a -> assertThat(a.getOwner()).isPositive())
                        .allSatisfy(a -> assertThat(a.getSpender()).isPositive())
                        .allSatisfy(a -> assertThat(a.getTokenId()).isPositive())
                        .allMatch(a -> recordItem.getConsensusTimestamp() == a.getTimestampLower())
                        .allMatch(a -> recordItem.getPayerAccountId().equals(a.getPayerAccountId())),
                () -> assertThat(nftRepository.findAll()).containsExactlyInAnyOrderElementsOf(expectedNfts),
                () -> assertThat(tokenAllowanceRepository.findAll())
                        .allSatisfy(a -> assertThat(a.getAmount()).isPositive())
                        .allSatisfy(a -> assertThat(a.getOwner()).isPositive())
                        .allSatisfy(a -> assertThat(a.getSpender()).isPositive())
                        .allSatisfy(a -> assertThat(a.getTokenId()).isPositive())
                        .allMatch(a -> recordItem.getConsensusTimestamp() == a.getTimestampLower())
                        .allMatch(a -> recordItem.getPayerAccountId().equals(a.getPayerAccountId()))
        );
    }

    private void assertCryptoTransaction(TransactionBody transactionBody, TransactionRecord record) {
        Entity actualAccount = getTransactionEntity(record.getConsensusTimestamp());
        assertAll(
                () -> assertTransactionAndRecord(transactionBody, record),
                () -> assertAccount(record.getReceipt().getAccountID(), actualAccount),
                () -> assertEntity(actualAccount));
    }

    private void assertCryptoEntity(CryptoCreateTransactionBody expected, Timestamp consensusTimestamp) {
        Entity actualAccount = getTransactionEntity(consensusTimestamp);
        long timestamp = DomainUtils.timestampInNanosMax(consensusTimestamp);
        assertAll(
                () -> assertEquals(expected.getAutoRenewPeriod().getSeconds(), actualAccount.getAutoRenewPeriod()),
                () -> assertEquals(timestamp, actualAccount.getCreatedTimestamp()),
                () -> assertEquals(false, actualAccount.getDeleted()),
                () -> assertNull(actualAccount.getExpirationTimestamp()),
                () -> assertArrayEquals(expected.getKey().toByteArray(), actualAccount.getKey()),
                () -> assertEquals(0, actualAccount.getMaxAutomaticTokenAssociations()),
                () -> assertEquals(expected.getMemo(), actualAccount.getMemo()),
                () -> assertEquals(timestamp, actualAccount.getTimestampLower()),
                () -> assertEquals(DomainUtils.getPublicKey(expected.getKey().toByteArray()),
                        actualAccount.getPublicKey()),
                () -> assertEquals(EntityId.of(expected.getProxyAccountID()),
                        actualAccount.getProxyAccountId()),
                () -> assertEquals(expected.getReceiverSigRequired(), actualAccount.getReceiverSigRequired())
        );
    }

    protected IterableAssert<CryptoTransfer> assertCryptoTransfers(
            int expectedNumberOfCryptoTransfers) {
        return assertThat(
                cryptoTransferRepository.findAll())
                .hasSize(expectedNumberOfCryptoTransfers)
                .allSatisfy(a -> assertThat(a.getId().getAmount()).isNotZero());
    }

    private List<NftAllowance> customizeNftAllowances(Timestamp consensusTimestamp, List<Nft> expectedNfts) {
        var delegatingSpender = recordItemBuilder.accountId();
        var owner = recordItemBuilder.accountId();
        var spender1 = recordItemBuilder.accountId();
        var spender2 = recordItemBuilder.accountId();
        var tokenId = recordItemBuilder.tokenId();
        var nft1 = Nft.builder()
                .id(new NftId(1L, EntityId.of(tokenId)))
                .accountId(EntityId.of(owner))
                .createdTimestamp(101L)
                .modifiedTimestamp(101L)
                .build();
        var nft2 = Nft.builder()
                .id(new NftId(2L, EntityId.of(tokenId)))
                .accountId(EntityId.of(owner))
                .createdTimestamp(102L)
                .modifiedTimestamp(102L)
                .build();
        var nft3 = Nft.builder()
                .id(new NftId(3L, EntityId.of(tokenId)))
                .accountId(EntityId.of(owner))
                .createdTimestamp(103L)
                .modifiedTimestamp(103L)
                .build();
        var timestamp = DomainUtils.timeStampInNanos(consensusTimestamp);
        List<NftAllowance> nftAllowances = new LinkedList<>();

        nftAllowances.add(NftAllowance.newBuilder()
                .setDelegatingSpender(delegatingSpender)
                .setOwner(owner)
                .addSerialNumbers(1L)
                .addSerialNumbers(2L)
                .setSpender(spender1)
                .setTokenId(tokenId)
                .build());
        expectedNfts.add(nft1.toBuilder()
                .delegatingSpender(EntityId.of(delegatingSpender))
                .modifiedTimestamp(timestamp)
                .spender(EntityId.of(spender1))
                .build());

        nftAllowances.add(NftAllowance.newBuilder()
                .setApprovedForAll(BoolValue.of(false))
                .setOwner(recordItemBuilder.accountId())
                .setSpender(recordItemBuilder.accountId())
                .setTokenId(recordItemBuilder.tokenId())
                .build());
        nftAllowances.add(NftAllowance.newBuilder()
                .setApprovedForAll(BoolValue.of(true))
                .setOwner(recordItemBuilder.accountId())
                .setSpender(recordItemBuilder.accountId())
                .setTokenId(recordItemBuilder.tokenId())
                .build());

        nftAllowances.add(NftAllowance.newBuilder()
                .setApprovedForAll(BoolValue.of(true))
                .setOwner(owner)
                .addSerialNumbers(2L)
                .addSerialNumbers(3L)
                .setSpender(spender2)
                .setTokenId(tokenId)
                .build());

        // duplicate nft allowance
        nftAllowances.add(nftAllowances.get(nftAllowances.size() - 1));

        // serial number 2's allowance is granted twice, the allowance should be granted to spender2 since it appears
        // after the nft allowance to spender1
        expectedNfts.add(nft2.toBuilder().modifiedTimestamp(timestamp).spender(EntityId.of(spender2)).build());
        expectedNfts.add(nft3.toBuilder().modifiedTimestamp(timestamp).spender(EntityId.of(spender2)).build());

        nftRepository.saveAll(List.of(nft1, nft2, nft3));

        return nftAllowances;
    }

    private void createAccount() {
        Transaction createTransaction = cryptoCreateTransaction();
        TransactionBody createTransactionBody = getTransactionBody(createTransaction);
        TransactionRecord createRecord = transactionRecordSuccess(createTransactionBody);
        parseRecordItemAndCommit(new RecordItem(createTransaction, createRecord));
    }

    private CryptoCreateTransactionBody.Builder cryptoCreateAccountBuilderWithDefaults() {
        return CryptoCreateTransactionBody.newBuilder()
                .setAutoRenewPeriod(Duration.newBuilder().setSeconds(1500L))
                .setInitialBalance(INITIAL_BALANCE)
                .setKey(keyFromString(KEY))
                .setMemo("CryptoCreateAccount memo")
                .setNewRealmAdminKey(keyFromString(KEY2))
                .setProxyAccountID(PROXY)
                .setRealmID(RealmID.newBuilder().setShardNum(0).setRealmNum(0).build())
                .setShardID(ShardID.newBuilder().setShardNum(0))
                .setReceiverSigRequired(true);
    }

    private Transaction cryptoCreateTransaction() {
        return cryptoCreateTransaction(cryptoCreateAccountBuilderWithDefaults());
    }

    private Transaction cryptoCreateTransaction(CryptoCreateTransactionBody.Builder cryptoCreateBuilder) {
        return buildTransaction(builder -> builder.setCryptoCreateAccount(cryptoCreateBuilder));
    }

    private Transaction cryptoAddLiveHashTransaction() {
        return buildTransaction(builder -> builder.getCryptoAddLiveHashBuilder()
                .getLiveHashBuilder()
                .setAccountId(accountId1)
                .setDuration(Duration.newBuilder().setSeconds(10000L))
                .setHash(ByteString.copyFromUtf8("live hash"))
                .setKeys(KeyList.newBuilder().addKeys(keyFromString(KEY))));
    }

    private Transaction cryptoDeleteLiveHashTransaction() {
        return buildTransaction(builder -> builder.getCryptoDeleteLiveHashBuilder()
                .setAccountOfLiveHash(accountId1)
                .setLiveHashToDelete(ByteString.copyFromUtf8("live hash")));
    }

    private Transaction cryptoDeleteTransaction() {
        return buildTransaction(builder -> builder.getCryptoDeleteBuilder()
                .setDeleteAccountID(accountId1));
    }

    private Transaction cryptoUpdateTransaction(AccountID accountNum) {
        return cryptoUpdateTransaction(accountNum, b -> {
        });
    }

    private Transaction cryptoUpdateTransaction(AccountID accountNum,
                                                Consumer<CryptoUpdateTransactionBody.Builder> custom) {
        return buildTransaction(builder -> {
            var cryptoBuilder = builder.getCryptoUpdateAccountBuilder()
                    .setAccountIDToUpdate(accountNum)
                    .setAutoRenewPeriod(Duration.newBuilder().setSeconds(1500L))
                    .setExpirationTime(Utility.instantToTimestamp(Instant.now()))
                    .setKey(keyFromString(KEY))
                    .setMaxAutomaticTokenAssociations(Int32Value.of(10))
                    .setMemo(StringValue.of("CryptoUpdateAccount memo"))
                    .setProxyAccountID(PROXY_UPDATE)
                    .setReceiverSigRequiredWrapper(BoolValue.of(false))
                    .setStakedNodeId(1L);
            custom.accept(cryptoBuilder);
        });
    }

    private Transaction cryptoTransferTransaction() {
        return buildTransaction(builder -> {
            for (int i = 0; i < additionalTransfers.length; i++) {
                builder.getCryptoTransferBuilder().getTransfersBuilder()
                        .addAccountAmounts(accountAmount(additionalTransfers[i], additionalTransferAmounts[i]));
            }
        });
    }

    private void groupCryptoTransfersByAccountId(TransactionRecord.Builder recordBuilder,
                                                 List<AccountAmount.Builder> amountsToBeAdded) {
        var accountAmounts = recordBuilder.getTransferListBuilder().getAccountAmountsBuilderList();

        var transfers = new HashMap<AccountID, Long>();
        Stream.concat(accountAmounts.stream(), amountsToBeAdded.stream())
                .forEach(accountAmount ->
                        transfers.compute(accountAmount.getAccountID(), (k, v) -> {
                            long currentValue = (v == null) ? 0 : v;
                            return currentValue + accountAmount.getAmount();
                        })
                );

        TransferList.Builder transferListBuilder = TransferList.newBuilder();
        transfers.entrySet().forEach(entry -> {
            AccountAmount accountAmount = AccountAmount.newBuilder().setAccountID(entry.getKey())
                    .setAmount(entry.getValue()).build();
            transferListBuilder.addAccountAmounts(accountAmount);
        });
        recordBuilder.setTransferList(transferListBuilder);
    }

    private void testRawBytes(Transaction transaction, byte[] expectedBytes) {
        // given
        TransactionBody transactionBody = getTransactionBody(transaction);
        TransactionRecord record = transactionRecordSuccess(transactionBody);

        // when
        parseRecordItemAndCommit(new RecordItem(transaction, record));

        // then
        var dbTransaction = getDbTransaction(record.getConsensusTimestamp());
        assertArrayEquals(expectedBytes, dbTransaction.getTransactionBytes());
    }

    private TransactionRecord transactionRecord(TransactionBody transactionBody, ResponseCodeEnum responseCode) {
        return transactionRecord(transactionBody, responseCode.getNumber(), recordBuilder -> {
        });
    }

    private TransactionRecord transactionRecord(TransactionBody transactionBody, int responseCode) {
        return transactionRecord(transactionBody, responseCode, recordBuilder -> {
        });
    }

    private TransactionRecord transactionRecord(TransactionBody transactionBody, int status,
                                                Consumer<TransactionRecord.Builder> builderConsumer) {
        return buildTransactionRecord(recordBuilder -> {
                    recordBuilder.getReceiptBuilder().setAccountID(accountId1);
                    builderConsumer.accept(recordBuilder);
                },
                transactionBody,
                status);
    }

    private TransactionRecord transactionRecordSuccess(TransactionBody transactionBody) {
        return transactionRecord(transactionBody, ResponseCodeEnum.SUCCESS);
    }

    private TransactionRecord transactionRecordSuccess(TransactionBody transactionBody,
                                                       Consumer<TransactionRecord.Builder> customBuilder) {
        return transactionRecord(transactionBody, ResponseCodeEnum.SUCCESS.getNumber(), customBuilder);
    }
}
